from pydantic import BaseModel, Field
import pathlib
from typing import Optional, Any, Union
import safer

from . import linear_reactions as lr
from . import point_reactions as pr
from . import geom_ops as geom


class LinearWallModel(BaseModel):
    height: float
    length: float
    vertical_spread_angle: float = 0.0
    minimum_point_spread: float = 0.5
    distributed_loads: dict = Field(default={})
    point_loads: dict = Field(default={})
    gravity_dir: str = "z"
    inplane_dir: str = "x"
    out_of_plane_dir: str = "y"
    apply_spread_angle_gravity: bool = True
    apply_spread_angle_inplane: bool = True
    magnitude_point_key: str = "p"
    magnitude_start_key: str = "w0"
    magnitude_end_key: str = "w1"
    location_point_key: str = "x"
    location_start_key: str = "x0"
    location_end_key: str = "x1"
    reverse_reaction_force_direction: bool = True
    _projected_loads: Optional[dict] = None
    """
    A model of a linear load-bearing wall segment. The segment is assumed to be linear
    and continuous.

    'height': The height of the wall
    'lenght': The length of the wall
    'vertical_spread_angle': Angle in degrees. Applied loads in the gravity direction
        spread through to the base of the wall according to this angle. The angle is 
        measured as deviation off of the direction of gravity (i.e. 0.0 is the in the
        gravity direction and 30.0 is 30 degrees away from vertical)
    'minimum_point_spread': For point loads applied without a spread angle, this is minimum width
        over which the point load will be spread. Whether this value is realistic or not depends on
        your context and the length unit system you are using. In other words, all applied point loads
        will be converted to distributed loads spread over this distance unless a vertical spread
        angle is applied.
    'distributed_loads': A dictionary of loads. Can be set directly or using the .add_dist_load()
        methods
    'point_loads': A dictionary of loads. Can be set directly or using the .add_point_load()
        methods
    'gravity_dir': A label used for the gravity direction. Must match the direction labels
        used in the applied loads for applied loads in the gravity direction.
    'inplane_dir': A label used for the inplane direction. Must match the direction labels
        used in the applied loads for applied loads in the inplane direction.
    'out_of_plane_dir': A label used for the out_of_plane direction. Must match the direction labels
        used in the applied loads for applied loads in the out_of_plane direction.
    'magnitude_point_key': The key that will be used internally to describe the magnitude of
        point loads.
    'magnitude_start_key': The key that will be used internally (and in reaction results) for the
        start magnitude
    'magnitude_end_key': The key that will be used internally (and in reaction results) for the
        end magnitude
    'location_key': The key that will be used internally to describe the location of applied
        point loads.
    'location_start_key': The key that will be used internally (and in reaction results) for the
        start location
    'location_end_key': The key that will be used internally (and in reaction results) for the
        end location
    'reverse_reaction_force_direction': If True, the reactions will be represented with a load
        of the opposite sign as the input load.
    """

    @classmethod
    def from_json(cls, filepath: str | pathlib.Path):
        with safer.open(filepath) as file:
            json_data = file.read()
        return cls.model_validate_json(json_data)

    def to_json(self, filepath: str | pathlib.Path, indent=2):
        json_data = self.model_dump_json(indent=indent)
        with safer.open(filepath, "w") as file:
            file.write(json_data)

    @classmethod
    def from_dict(cls, data: dict):
        return cls.model_validate(data)

    def dump_dict(self):
        return self.model_dump(mode="json")

    def add_dist_load(
        self,
        magnitude_start: float,
        magnitude_end: float,
        location_start: float,
        location_end: float,
        case: str,
        dir: str,
    ) -> None:
        """
        Adds a distributed load to the model

        'magnitude_start': The magnitude at the start of the linear distributed load
        'magnitude_end': The magnitude at the end ...
        'location_start': The start location on the wall of the dist load (>=0.0)
        'location_end': The end location on the wall of the dist load (<= self.length)
        'case': The load case of this load
        'dir': The direction the load is to be applied
        """
        assert location_start >= 0.0
        assert location_end <= self.length

        self.distributed_loads.setdefault(dir, {})
        self.distributed_loads[dir].setdefault(case, [])
        self.distributed_loads[dir][case].append(
            {
                self.magnitude_start_key: magnitude_start,
                self.magnitude_end_key: magnitude_end,
                self.location_start_key: location_start,
                self.location_end_key: location_end,
            }
        )

    def add_point_load(
        self, magnitude: float, location: float, case: str, dir: str
    ) -> None:
        """
        Adds a point load to the model

        'magnitude_start': The magnitude of the point load
        'location_start': The location on the wall of the point load (0.0 <= location <= self.length)
        'case': The load case of this load
        'dir': The direction the load is to be applied
        """
        assert 0.0 <= location <= self.length

        self.point_loads.setdefault(dir, {})
        self.point_loads[dir].setdefault(case, [])
        self.point_loads[dir][case].append(
            {
                self.magnitude_point_key: magnitude,
                self.location_point_key: location,
            }
        )

    def spread_loads(self) -> None:
        """
        Populates self._projected_loads with the loads projected from the distributed
        and point loads.

        This is an intermediate step to retrieving the consolidated reactions.
        """
        self._projected_loads = {}
        proj = self._projected_loads
        w0 = self.magnitude_start_key
        w1 = self.magnitude_end_key
        x0 = self.location_start_key
        x1 = self.location_end_key
        p = self.magnitude_point_key
        x = self.location_point_key
        for load_dir, load_cases in self.distributed_loads.items():
            proj.setdefault(load_dir, {})
            should_apply_spread_angle = (
                load_dir == self.gravity_dir
                and self.apply_spread_angle_gravity
                and self.vertical_spread_angle != 0.0
            ) or (
                load_dir == self.inplane_dir
                and self.apply_spread_angle_inplane
                and self.vertical_spread_angle != 0.0
            )
            for load_case, dist_loads in load_cases.items():
                proj[load_dir].setdefault(load_case, [])
                for dist_load in dist_loads:
                    if should_apply_spread_angle:
                        projected_load = geom.apply_spread_angle(
                            self.height,
                            self.length,
                            self.vertical_spread_angle,
                            w0=dist_load[w0],
                            x0=dist_load[x0],
                            w1=dist_load.get(w1),
                            x1=dist_load.get(x1),
                        )
                        proj[load_dir][load_case].append(
                            {
                                w0: projected_load[0],
                                w1: projected_load[1],
                                x0: projected_load[2],
                                x1: projected_load[3],
                            }
                        )
                    else:
                        proj[load_dir][load_case].append(dist_load)

        for load_dir, load_cases in self.point_loads.items():
            proj.setdefault(load_dir, {})
            should_apply_spread_angle = (
                load_dir == self.gravity_dir
                and self.apply_spread_angle_gravity
                and self.vertical_spread_angle != 0.0
            ) or (
                load_dir == self.inplane_dir
                and self.apply_spread_angle_inplane
                and self.vertical_spread_angle != 0.0
            )
            for load_case, point_loads in load_cases.items():
                proj[load_dir].setdefault(load_case, [])
                for point_load in point_loads:
                    if should_apply_spread_angle:
                        projected_load = geom.apply_spread_angle(
                            self.height,
                            self.length,
                            self.vertical_spread_angle,
                            p=point_load[p],
                            x=point_load[x],
                        )
                        proj[load_dir][load_case].append(
                            {
                                w0: projected_load[0],
                                w1: projected_load[1],
                                x0: projected_load[2],
                                x1: projected_load[3],
                            }
                        )
                    else:
                        projected_load = geom.apply_minimum_width(
                            point_load[p],
                            point_load[x],
                            self.minimum_point_spread,
                            self.length,
                        )
                        proj[load_dir][load_case].append(
                            {
                                w0: projected_load[0],
                                w1: projected_load[1],
                                x0: projected_load[2],
                                x1: projected_load[3],
                            }
                        )

        self._projected_loads = proj

    def get_reactions(
        self,
        flattened: bool = False,
        direction_key: str = "dir",
        case_key: str = "case",
    ):
        self.spread_loads()  # Populates self._projected_loads
        lrs = lr.LinearReactionString.from_projected_loads(
            self._projected_loads,
            self.magnitude_start_key,
            self.magnitude_end_key,
            self.location_start_key,
            self.location_end_key,
            self.reverse_reaction_force_direction,
        )
        return lrs.consolidate_reactions(
            flatten=flattened, dir_key=direction_key, case_key=case_key
        )



from dataclasses import dataclass


@dataclass
class PointReactionCollection:
    _reaction_components: dict[str, dict[str, list[float]]]

    def extract_reactions(self, case: str, dir: str) -> float:
        """
        Returns a float representing the total sum of loads for the
        given 'dir' (direction) and 'case' (load case).
        """
        return sum(self._reaction_components.get(case, {}).get(dir, []))

    @classmethod
    def from_point_loads(cls, point_loads: dict[str, dict[str, str | float]]):
        """ """
        reaction_components = {}
        for load_case in point_loads.items():
            reaction_components.setdefault(load_case, {})
            for load_dir, pt_loads in load_case.items():
                reaction_components[load_case].setdefault(load_dir, [])
                for point_load in pt_loads:
                    reaction_components[load_dir][load_case].append(
                        point_load["magnitude"]
                    )
        return cls(reaction_components)



from dataclasses import dataclass
from typing import Optional
from load_distribution import Singularity, singularities_to_polygon
from .geom_ops import round_to_close_integer as rtci


@dataclass
class LinearReaction:
    w0: float
    w1: float
    x0: float
    x1: float

    def point_in_reaction(self, x: float):
        return self.x0 <= x <= self.x1

    def points_enclose_reaction(self, xa: float, xb: float) -> bool:
        """
        Returns True if xa <= self.x0 <= self.x1 <= xb
        """
        return xa <= self.x0 <= self.x1 <= xb

    def extract_reaction(self, xa: float, xb: float) -> "LinearReaction":
        """
        Returns the portion of the reaction that would exist between
        'xa' and 'xb'
        """
        m = (self.w1 - self.w0) / (self.x1 - self.x0)
        y0 = self.w0
        if not any(
            [
                self.point_in_reaction(xa),
                self.point_in_reaction(xb),
                self.points_enclose_reaction(xa, xb),
            ]
        ):
            return LinearReaction(0.0, 0.0, self.x0, self.x1)
        if not self.point_in_reaction(xa):
            xi = self.x0
            yi = self.w0
        else:
            xi = xa
            yi = y0 + (xi - self.x0) * m

        if not self.point_in_reaction(xb):
            xj = self.x1
            yj = self.w1
        else:
            xj = xb
            yj = y0 + (xj - self.x0) * m

        return LinearReaction(rtci(yi), rtci(yj), rtci(xi), rtci(xj))


@dataclass
class LinearReactionString:
    """
    A class to manage a collection of LinearReactions
    """

    linear_reactions: dict[str, dict[str, list[LinearReaction]]]
    magnitude_start_key: str
    magnitude_end_key: str
    location_start_key: str
    location_end_key: str
    reverse_reaction_direction: bool = True

    @classmethod
    def from_projected_loads(
        cls,
        projected_loads: dict[str, dict[str, list[dict]]],
        magnitude_start_key: str,
        magnitude_end_key: str,
        location_start_key: str,
        location_end_key: str,
        reverse_reaction_direction: bool = True,
    ):
        w0 = magnitude_start_key
        w1 = magnitude_end_key
        x0 = location_start_key
        x1 = location_end_key
        reverse_direction = -1 if reverse_reaction_direction else 1
        linear_reaction_components = {}
        for load_dir, load_cases in projected_loads.items():
            linear_reaction_components.setdefault(load_dir, {})
            for load_case, applied_loads in load_cases.items():
                linear_reaction_components[load_dir].setdefault(load_case, [])
                for applied_load in applied_loads:
                    linear_reaction = LinearReaction(
                        applied_load[w0] * reverse_direction,
                        applied_load.get(w1) * reverse_direction,
                        applied_load[x0],
                        applied_load.get(x1),
                    )

                    linear_reaction_components[load_dir][load_case].append(
                        linear_reaction
                    )
        return cls(
            linear_reaction_components, w0, w1, x0, x1, reverse_reaction_direction
        )

    def extract_reaction_string(self, xa: float, xb: float, case: str, dir: str):
        """
        Returns a LinearReactionString representing the linear reactions that
        exist between 'xa' and 'xb' extracted from self.

        Returns None if there are no reactions within the 'x0' and 'x1' extents
        """
        extracted = {}
        extracted.setdefault(dir, {})
        extracted[dir].setdefault(case, [])
        for reaction in self.linear_reactions.get(dir, {}).get(case, {}):
            extracted_linear_reaction = reaction.extract_reaction(xa, xb)
            if extracted_linear_reaction.w0 != 0 and extracted_linear_reaction.w1 != 0:
                extracted[dir][case].append()
        return LinearReactionString(
            extracted,
            self.magnitude_start_key,
            self.magnitude_end_key,
            self.location_start_key,
            self.location_end_key,
        )

    def consolidate_reactions(
        self, flatten: bool, dir_key: str = "dir", case_key: str = "case"
    ):
        """
        Collects distributed loads from the top of a wall run and
        converts them into a LinearReactionString which can sum and
        parcel out the reactions into wall segments or beams that are
        supporting the wall run.

        A dict of 'dist_loads' should be organized as follows:

            {
                "dir1": {
                    "lc": [
                        {"w0": float, "w1": float, "x0": float, "x1": float}
                    ],
                    ...
                },
                ...
            }

        If 'flatten' is True then the results will be a list of load dicts.
            In this case then 'dir_key' and 'case_key' will be used to embed
            the direction and load case into each load_dict.
            Otherwise, the result will be a tree nested by direction and then
            by load case.
        """
        w0 = self.magnitude_start_key
        w1 = self.magnitude_end_key
        x0 = self.location_start_key
        x1 = self.location_end_key
        reaction_components = {}
        flattened_reaction_components = []
        for load_dir, load_cases in self.linear_reactions.items():
            reaction_components.setdefault(load_dir, {})
            for load_case, linear_reactions in load_cases.items():
                reaction_components[load_dir].setdefault(load_case, [])
                singularity_functions = []
                for lr in linear_reactions:
                    if lr.w1 is None and lr.x1 is None:
                        point_load = {
                            w0: lr.w0,
                            x0: lr.x0,
                            dir_key: load_dir,
                            case_key: load_case,
                        }
                        flattened_reaction_components.append(point_load)
                        reaction_components[load_dir][load_case].append(point_load)
                    else:
                        m = (lr.w1 - lr.w0) / (lr.x1 - lr.x0)
                        y0 = lr.w0
                        singularity_function = Singularity(
                            x0=lr.x0, y0=y0, x1=lr.x1, m=m, precision=6
                        )
                        singularity_functions.append(singularity_function)
                if not singularity_functions:
                    continue
                linear_reactions = singularity_xy_to_distributed_loads(
                    singularities_to_polygon(singularity_functions, xy=True),
                    magnitude_start_key=w0,
                    magnitude_end_key=w1,
                    location_start_key=x0,
                    location_end_key=x1,
                    case=load_case,
                    dir=load_dir,
                    case_key="case",
                    dir_key="dir",
                )
                flattened_reaction_components.append(linear_reactions)

                # Get ride of the extrandious dir and case keys for unflattened results
                reaction_components[load_dir][load_case] += linear_reactions
        if flatten:
            return flattened_reaction_components
        return reaction_components


def filter_repeated_y_values(
    xy_vals: list[list[float], list[float]],
) -> list[list[float, float]]:
    """
    Returns xy_vals but with any "repeating" data points removed and
    returns a list of coordinates, list[list[float, float]]
    """
    coords = list(zip(*xy_vals))
    filtered = []
    for idx, (x, y) in enumerate(coords):
        next_y_idx = min(idx + 1, len(coords) - 1)
        next_y = coords[next_y_idx][1]
        if idx == 0:
            filtered.append([x, y])
            prev_y = y
        else:
            if prev_y == y == next_y:
                continue
            else:
                filtered.append([x, y])
        prev_y = y
    return filtered


def singularity_xy_to_distributed_loads(
    xy_vals: list[list[float], list[float]],
    magnitude_start_key: str,
    magnitude_end_key: str,
    location_start_key: str,
    location_end_key: str,
    case: str,
    dir: str,
    case_key: str = "case",
    dir_key: str = "dir",
) -> list[dict]:
    """
    Returns dicts representing distributed
    """
    w0 = magnitude_start_key
    w1 = magnitude_end_key
    x0 = location_start_key
    x1 = location_end_key
    dist_loads = []
    prev_x = None
    for idx, (x, y) in enumerate(zip(*xy_vals)):
        if idx == 0:
            continue
        if prev_x is None:
            prev_x = x
            prev_y = y
        elif x - prev_x > 1e-3:
            dist_load = {
                w0: float(rtci(prev_y)),
                w1: float(rtci(y)),
                x0: float(rtci(prev_x)),
                x1: float(rtci(x)),
                case_key: case,
                dir_key: dir,
            }
            dist_loads.append(dist_load)
            prev_x = x
            prev_y = y
        else:
            prev_x = x
            prev_y = y
    return dist_loads



import math
from typing import Optional


def apply_spread_angle(
    wall_height: float,
    wall_length: float,
    spread_angle: float,
    w0: Optional[float] = None,
    x0: Optional[float] = None,
    w1: Optional[float] = None,
    x1: Optional[float] = None,
    p: Optional[float] = None,
    x: Optional[float] = None,
) -> tuple[float, float, float, float]:
    """
    Returns a dictionary representing the load described by
    w0, w1, x0, x1 (if distributed load) or p, x (if point
    load).

    The total spread cannot be longer than the wall length.

    spread_angle is assumed to be in degrees
    """
    angle_rads = math.radians(spread_angle)
    spread_amount = wall_height * math.tan(angle_rads)
    if None not in [w0, w1, x0, x1]:
        projected_x0 = max(0.0, x0 - spread_amount)
        projected_x1 = min(wall_length, x1 + spread_amount)
        original_length = x1 - x0
    elif None not in [x, p]:
        projected_x0 = max(0.0, x - spread_amount)
        projected_x1 = min(wall_length, x + spread_amount)
        original_length = 0
    else:
        print(f"Weird condition: {locals()=}")

    projected_length = projected_x1 - projected_x0
    ratio = original_length / projected_length

    if None not in [w0, w1, x0, x1]:
        projected_w0 = w0 * ratio
        projected_w1 = w1 * ratio
    elif None not in [x, p]:
        projected_w0 = p / projected_length
        projected_w1 = p / projected_length
    return (
        round_to_close_integer(projected_w0),
        round_to_close_integer(projected_w1),
        round_to_close_integer(projected_x0),
        round_to_close_integer(projected_x1),
    )


def apply_minimum_width(
    magnitude: float,
    location: float,
    spread_width: float,
    wall_length: float,
) -> tuple[float, float, float, float]:
    """
    Returns a dictionary representing a distributed load
    representing the point load converted to a distributed
    load over the 'spread_width' in such a way that the
    point load will be distributed an equal amount over
    half of the spread_width on each side of point load.

    If the point load location is 0 or wall_length, then
    the point load will be a distributed load over half
    of the spread_width (since there is not room for the
    other half).

    Load locations between zero/wall_length and half of the
    spread_width will be linearly interpolated.
    """
    if spread_width > wall_length:
        spread_width = wall_length
    if location <= spread_width / 2:
        projected_x0 = 0
        projected_x1 = location + spread_width / 2
    elif (wall_length - location) <= spread_width / 2:
        projected_x0 = location - spread_width / 2
        projected_x1 = wall_length
    else:
        projected_x0 = location - spread_width / 2
        projected_x1 = location + spread_width / 2

    projected_x1 = location + spread_width / 2
    projected_w0 = projected_w1 = magnitude / (projected_x1 - projected_x0)
    return (
        round_to_close_integer(projected_w0),
        round_to_close_integer(projected_w1),
        round_to_close_integer(projected_x0),
        round_to_close_integer(projected_x1),
    )


def round_to_close_integer(x: float, eps=1e-7) -> float | int:
    """
    Rounds to the nearest int if it is REALLY close
    """
    if abs(abs(round(x)) - abs(x)) < eps:
        return round(x)
    else:
        return x



"""
A package that models a generic load-bearing wall for the purpose of determining
reactions and load-transfer.

The wall model is parameterizable and can represent any material
"""

__version__ = "0.4.1"

from loadbearing_wall.wall_model import LinearWallModel
from loadbearing_wall import *
